package org.arkerthan.sanityCheck;

import org.arkerthan.sanityCheck.element.AngleBracketElement;
import org.arkerthan.sanityCheck.element.CommentElement;
import org.arkerthan.sanityCheck.element.Element;
import org.arkerthan.sanityCheck.element.KnownElement;

import java.io.*;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;

/**
 * @author Arkerthan
 * @version 1.0
 */
public class Main {

	public static TagSearchTree<Tag> htmlTags, twineTags;
	private static String currentFile;
	private static int currentLine, currentPosition;
	private static Stack<Element> stack;
	private static List<SyntaxError> errors = new LinkedList<>();
	private static String[] excluded;

	public static void main(String[] args) {

		//setup
		setupExclude();
		setupHtmlTags();
		setupTwineTags();
		Path workingDir = Paths.get("").toAbsolutePath();

		//actual sanityCheck
		runSanityCheckInDirectory(workingDir, new File("src/"));

		//output errors
		for (SyntaxError e :
				errors) {
			System.out.println(e.getError());
		}
	}


	/**
	 * Goes through the whole directory including subdirectories and runs
	 * {@link Main#sanityCheck(Path)} on all .tw files
	 *
	 * @param dir to be checked
	 */
	private static void runSanityCheckInDirectory(Path workingDir, File dir) {
		//subdirectories are checked recursively

		try {
			for (File file : dir.listFiles()) {
				if (file.isFile()) { //run sanityCheck if file is a .tw file
					String path = file.getAbsolutePath();
					if (path.endsWith(".tw")) {
						sanityCheck(workingDir.relativize(file.toPath()));
					}
				} else if (file.isDirectory()) {
					runSanityCheckInDirectory(workingDir, file.getAbsoluteFile());
				}
			}
		} catch (NullPointerException e) {
			e.printStackTrace();
			System.err.println("Couldn't read directory " + currentFile);
			System.exit(-1);
		}
	}

	/**
	 * Runs the sanity check for one file. Does not run if file is excluded.
	 *
	 * @param path file to be checked
	 */
	private static void sanityCheck(Path path) {
		File file = path.toFile();

		// replace this with a known encoding if possible
		Charset encoding = Charset.defaultCharset();

		if (!excluded(file.getPath())) {
			currentFile = file.getPath();
			currentLine = 1;
			stack = new Stack<>();

			//actually opening and reading the file
			try (InputStream in = new FileInputStream(file);
				 Reader reader = new InputStreamReader(in, encoding);
				 // buffer for efficiency
				 Reader buffer = new BufferedReader(reader)) {
				handleCharacters(buffer);
			} catch (IOException e) {
				e.printStackTrace();
				System.err.println("Couldn't read " + file);
			}
		}
	}

	/**
	 * sets up a {@link TagSearchTree<Tag>} for fast access of HTML tags later
	 */
	private static void setupHtmlTags() {
		//load HTML tags into a list
		List<Tag> TagsList = loadTags("devTools/javaSanityCheck/htmlTags");

		//turn List into alphabetical search tree
		try {
			htmlTags = new TagSearchTree<>(TagsList);
		} catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("Illegal Character in devTools/javaSanityCheck/htmlTags");
			System.exit(-1);
		}
	}

	/**
	 * sets up a {@link TagSearchTree<Tag>} for fast access of twine tags later
	 */
	private static void setupTwineTags() {
		//load twine tags into a list
		List tagsList = loadTags("devTools/javaSanityCheck/twineTags");

		//turn List into alphabetical search tree
		try {
			twineTags = new TagSearchTree<>(tagsList);
		} catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("Illegal Character in devTools/javaSanityCheck/twineTags");
			System.exit(-1);
		}
	}

	/**
	 * Loads a list of tags from a file
	 *
	 * @param filePath file to load tags from
	 * @return loaded tags
	 */
	private static List<Tag> loadTags(String filePath) {
		List<Tag> tagsList = new LinkedList<>();
		try {
			Files.lines(new File(filePath).toPath()).map(String::trim)
					.filter(s -> !s.startsWith("#"))
					.forEach(s -> tagsList.add(parseTag(s)));
		} catch (IOException e) {
			System.err.println("Couldn't read " + filePath);
		}
		return tagsList;
	}

	/**
	 * Turns a string into a Tag
	 * ";1" at the end of the String indicates that the tag needs to be closed later
	 *
	 * @param s tag as String
	 * @return tag as Tag
	 */
	private static Tag parseTag(String s) {
		String[] st = s.split(";");
		if (st.length > 1 && st[1].equals("1")) {
			return new Tag(st[0], false);
		}
		return new Tag(st[0], true);
	}

	/**
	 * sets up the excluded files array.
	 */
	private static void setupExclude() {
		//load excluded files
		List<String> excludedList = new ArrayList<>();
		try {
			Files.lines(new File("devTools/javaSanityCheck/excluded").toPath()).map(String::trim)
					.filter(s -> !s.startsWith("#"))
					.forEach(excludedList::add);
		} catch (IOException e) {
			System.err.println("Couldn't read devTools/javaSanityCheck/excluded");
		}

		//turn excluded files into an array and change path to windows style if needed
		if (isWindows()) {
			excluded = new String[excludedList.size()];
			int i = 0;
			for (String s :
					excludedList) {
				excluded[i++] = s.replaceAll("/", "\\\\");
			}
		} else {
			excluded = excludedList.toArray(new String[0]);
		}
	}

	/**
	 * @return whether OS is Windows or not
	 */
	private static boolean isWindows() {
		return (System.getProperty("os.name").toLowerCase(Locale.ENGLISH).contains("windows"));
	}

	/**
	 * checks if a file or directory is excluded from the sanity check
	 *
	 * @param s file/directory to be checked
	 * @return whether it is excluded or not
	 */
	private static boolean excluded(String s) {
		for (String ex :
				excluded) {
			if (s.startsWith(ex)) return true;
		}
		return false;
	}

	/**
	 * Reads the file character by character.
	 *
	 * @param reader reader that is read
	 * @throws IOException thrown if the file can't be read
	 */
	private static void handleCharacters(Reader reader) throws IOException {
		int r;
		while ((r = reader.read()) != -1) {
			char c = (char) r;
			handleCharacter(c);
		}
	}

	/**
	 * Handles a single character
	 *
	 * @param c next character
	 */
	private static void handleCharacter(char c) {
		//updating position
		currentPosition++;
		if (c == '\n') {
			currentLine++;
			currentPosition = 1;
		}

		//try applying to the innermost element
		if (!stack.empty()) {
			int change;
			try {
				change = stack.peek().handleChar(c);
			} catch (SyntaxError e) {
				change = e.getChange();
				addError(e);
			}

			//change greater 0 means the innermost element did some work
			if (change > 0) {
				//2 means the Element is complete
				if (change == 2) {
					//remove the topmost element from stack since it is complete
					stack.pop();
					return;
				}
				//3 means the Element is complete and part of a two or more tag system
				if (change == 3) {
					//remove the topmost element from stack since it is complete
					KnownElement k = stack.pop().getKnownElement();
					//if KnownElement k is closing another element, check if there is one and remove it
					if (k.isClosing()) {
						if (stack.empty()) { //there are no open elements at all
							addError(new SyntaxError("Closed tag " + k.getShortDescription() + " without " +
									"having any open tags.", -2));
						} else if (stack.peek() instanceof KnownElement) {
							//get opening tag
							KnownElement kFirst = (KnownElement) stack.pop();
							//check if closing element matches the opening element
							if (!kFirst.isMatchingElement(k)) {
								addError(new SyntaxError("Opening tag " + kFirst.getShortDescription() +
										" does not match closing tag " + k.getShortDescription() + ".", -2));
							}
						} else {
							//There closing tag inside another not Known element: <div </html>
							addError(new SyntaxError("Closing tag " + k.getShortDescription() + " inside " +
									"another tag " + stack.peek().getShortDescription() + " without opening first.",
									-2, true));
						}
					}
					//check if the element needs to be closed by another
					if (k.isOpening()) {
						stack.push(k);
					}
					return;
				}
				//means the element couldn't do anything with it and is finished
				if (change == 4) {
					stack.pop();
				} else {
					return;
				}
			}
		}


		//innermost element was uninterested, trying to find matching element
		switch (c) {
			//case '@':
			//stack.push(new AtElement(currentLine, currentPosition));
			//break;
			case '<':
				stack.push(new AngleBracketElement(currentLine, currentPosition));
				break;
			//case '>':
			//addError(new SyntaxError("Dangling \">\", current innermost: " + (stack.empty() ? "null" : stack.peek().getShortDescription()), -2));
			//break;
			case '/':
				stack.push(new CommentElement(currentLine, currentPosition));
				break;
			//case '(':
			//	stack.push(new BracketElement(currentLine, currentPosition));
		}
	}

	/**
	 * add an error to the error list
	 *
	 * @param e new error
	 */
	private static void addError(SyntaxError e) {
		e.setFile(currentFile);
		e.setLine(currentLine);
		e.setPosition(currentPosition);
		errors.add(e);
	}
}
